import argparse
from typing import Dict

from rich.markdown import Markdown

from .. import config
from . import BaseCommand


class Command(BaseCommand):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--all",
            action="store_true",
            help=(
                "By default, only configurations having a written "
                "description are shown.  If you would like to see "
                "all configurations, including those that are intended "
                "to be used as includes, use this flag."
            ),
        )
        return super().add_arguments(parser)

    def handle(self) -> None:
        to_show: Dict[str, config.Config] = {}
        all_configs = config.get_available_configs()

        for config_name in all_configs:
            conf = config.get_config_by_name(config_name)

            if self.options.all or conf.description:
                to_show[config_name] = conf

        if len(to_show) != len(all_configs):
            self.console.print(
                f"Showing {len(to_show)} of {len(all_configs)} configs; "
                "use --all to see more.",
                style="red",
            )

        for config_name, conf in to_show.items():
            formatted = Markdown(conf.description or "")

            self.console.print(f"[blue][b]{config_name}[/b][/blue]")
            if formatted:
                self.console.print(formatted, style="italic")
            if conf.alignment_holes:
                self.console.print(f"- Alignment Holes")
            if conf.isolation_routing:
                self.console.print("- Isolation Routing")
            if conf.edge_cuts:
                self.console.print("- Edge Cuts")
            if conf.drill:
                self.console.print("- Drill Profiles")
                for k, v in conf.drill.items():
                    self.console.print(
                        f"  - {k}: {v.min_size or '0'}-{v.max_size or 'Infinity'}"
                    )



import argparse
import os

import yaml

from .. import config
from . import BaseCommand


class Command(BaseCommand):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "name", help="Name of the configuration you would like to generate."
        )
        return super().add_arguments(parser)

    def handle(self) -> None:
        final_path = f"{self.options.name}.yaml"

        os.makedirs(
            os.path.dirname(config.get_user_config_path(final_path)), exist_ok=True
        )
        with open(config.get_user_config_path(final_path), "w") as outf:
            outf.write(yaml.safe_dump(config.get_default_config_dict()))

        self.console.print(
            f"Configuration '{self.options.name}' written to '{final_path}'."
        )



import argparse
from barbari.exceptions import BarbariFlatcamError
import subprocess

from .. import config
from . import BaseCommand


class Command(BaseCommand):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--flatcam",
            # default="./FlatCAM.py",
            required=True,
            help="Path to flatcam executable (FlatCAM.py)",
        )
        parser.add_argument(
            "--python-bin",
            default="python",
            help=(
                "Path to the python binary to use when running "
                "FlatCAM.py; set this to the correct python "
                "binary for a virtualenvironment if you are "
                "using one."
            ),
        )
        return super().add_arguments(parser)

    def handle(self) -> None:
        proc = subprocess.Popen(
            [
                self.options.python_bin,
                self.options.flatcam,
                "--help",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = proc.communicate()
        if "--shellfile" not in stdout.decode("utf-8"):
            raise BarbariFlatcamError(
                "Could not start FlatCam using the provided parameters: "
                f"stdout: {stdout}, stderr: {stderr}"
            )

        self.config.flatcam_path = self.options.flatcam
        self.config.python_bin = self.options.python_bin

        config.save_environment_config(self.config)
        self.console.print("[green]Flatcam found. [b]Configuration saved[/b][/green]")



from __future__ import annotations

import copy
from dataclasses import dataclass, asdict
import logging
import os
from typing import cast, Dict, Iterable, List, Optional, Type, Union, Tuple

import appdirs
import yaml

from . import exceptions


logger = logging.getLogger(__name__)


class JobSpec(object):
    def __init__(self, data, name=None):
        self._data = data
        self._name = name

        super().__init__()

    @property
    def name(self) -> str:
        return self._name

    @property
    def tool_size(self) -> float:
        return self._data["tool_size"]

    @property
    def cut_z(self) -> float:
        return self._data["cut_z"]

    @property
    def travel_z(self) -> float:
        return self._data["travel_z"]

    @property
    def feed_rate(self) -> float:
        return self._data["feed_rate"]

    @property
    def spindle_speed(self) -> int:
        return self._data["spindle_speed"]

    @property
    def multi_depth(self) -> bool:
        return self._data.get("multi_depth", False)

    @property
    def depth_per_pass(self) -> float:
        return self._data.get("depth_per_pass")

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name or self._data}>"


class MillHolesJobSpec(JobSpec):
    pass


class MillSlotsJobSpec(MillHolesJobSpec):
    pass


class IsolationRoutingJobSpec(JobSpec):
    @property
    def passes(self) -> float:
        return self._data["passes"]

    @property
    def pass_overlap(self) -> float:
        return self._data.get("pass_overlap", 1.0)


class BoardCutoutJobSpec(JobSpec):
    @property
    def margin(self) -> float:
        return self._data["margin"]

    @property
    def gap_size(self) -> float:
        return self._data["gap_size"]

    @property
    def gaps(self) -> str:
        return self._data["gaps"]


class DrillHolesJobSpec(JobSpec):
    @property
    def drill_z(self) -> float:
        return self._data.get("drill_z")


class ToolProfileSpec(JobSpec):
    @property
    def min_size(self) -> float:
        return self._data.get("min_size", 0)

    @property
    def max_size(self) -> float:
        return self._data.get("max_size", 999)

    @property
    def range(self) -> Tuple[float, float]:
        return self.min_size, self.max_size

    @property
    def range_center(self) -> float:
        return self.min_size + ((self.max_size - self.min_size) / 2)

    def allowed_for_tool_size(self, diameter: float) -> bool:
        range_allowed = "min_size" in self._data or "max_size" in self._data

        if (
            range_allowed and (self.min_size <= diameter <= self.max_size)
        ) or diameter in self.sizes:
            return True

        return False

    def is_better_match_than(
        self, diameter: float, other: Optional[ToolProfileSpec]
    ) -> bool:
        if not other:
            return True

        # If one spec specifically mentions a particular drill
        # size, it wins.
        if diameter in self.sizes and diameter not in other.sizes:
            return True
        elif diameter in other.sizes and diameter not in self.sizes:
            return False

        # If one spec is using only drilling profiles, and matches
        # our hole size exactly, that one wins
        if (
            all(isinstance(spec, DrillHolesJobSpec) for spec in self.specs)
            and max(spec.tool_size <= diameter for spec in self.specs) == diameter
        ):
            return True
        elif (
            all(isinstance(spec, DrillHolesJobSpec) for spec in other.specs)
            and max(spec.tool_size <= diameter for spec in other.specs) == diameter
        ):
            return False

        # If one spec is all milling profiles with a tool size below
        # our diameter, that side wins
        if all(isinstance(spec, MillHolesJobSpec) for spec in self.specs) and all(
            spec.tool_size <= diameter for spec in self.specs
        ):
            return True
        elif all(isinstance(spec, MillHolesJobSpec) for spec in other.specs) and all(
            spec.tool_size <= diameter for spec in other.specs
        ):
            return False

        if abs(diameter - self.range_center) < abs(diameter - other.range_center):
            return True
        elif abs(diameter - other.range_center) < abs(diameter - self.range_center):
            return False

        return False

    @property
    def sizes(self) -> List[float]:
        return self._data.get("sizes", [])

    @property
    def specs(self) -> Iterable[Union[MillHolesJobSpec, DrillHolesJobSpec]]:
        specs = []

        for spec_data in self._data["specs"]:
            data = spec_data["params"]
            spec_type = spec_data["type"]
            spec_class: Union[Type[MillHolesJobSpec], Type[DrillHolesJobSpec]]

            if spec_type == "cnc_drill":
                spec_class = DrillHolesJobSpec
            elif spec_type == "mill_holes":
                spec_class = MillHolesJobSpec
            elif spec_type == "mill_slots":
                spec_class = MillSlotsJobSpec
            else:
                raise exceptions.InvalidConfiguration(
                    "Unexpected spec type: %s" % spec_type
                )

            specs.append(spec_class(data))

        return specs


class DrillProfileSpec(ToolProfileSpec):
    @property
    def specs(self) -> Iterable[Union[MillHolesJobSpec, DrillHolesJobSpec]]:
        specs = super().specs

        for spec in specs:
            if not isinstance(spec, (DrillHolesJobSpec, MillHolesJobSpec)):
                raise exceptions.InvalidConfiguration(
                    "Drills support only 'cnc_drill' and 'mill_holes' specifications."
                )

        return specs


class SlotProfileSpec(ToolProfileSpec):
    @property
    def specs(self) -> Iterable[MillSlotsJobSpec]:
        specs = super().specs

        for spec in specs:
            if not isinstance(spec, MillSlotsJobSpec):
                raise exceptions.InvalidConfiguration(
                    "Slots support only 'mill_slots' specifications."
                )

        yield from cast(Iterable[MillSlotsJobSpec], specs)


class AlignmentHolesJobSpec(MillHolesJobSpec):
    @property
    def mirror_axis(self) -> str:
        return self._data.get("mirror_axis", "X")

    @property
    def hole_size(self) -> float:
        return self._data["hole_size"]

    @property
    def hole_offset(self) -> float:
        return self._data["hole_offset"]


class Config(object):
    def __init__(self, data):
        self._data = data

    @classmethod
    def from_file(self, path) -> Config:
        configs: List[Config] = []

        with open(path, "r") as inf:
            loaded = yaml.safe_load(inf)

        includes = loaded.pop("include", [])

        configs.append(Config(loaded))

        config_dir = os.path.dirname(path)
        for include in includes:
            if os.path.splitext(include)[1] in (".yaml", ".yml"):
                include_path = os.path.join(config_dir, include)

                configs.append(Config.from_file(include_path))
            else:
                configs.append(get_config_by_name(include))

        merged = sum(configs, Config({}))

        # By default, we strip descriptions when merging multiple configs;
        # but that's just because we can't make that sane when a user is
        # merging configs at the command-line on an ad-hoc basis; in this
        # particular case, the loaded config *does* know what files are
        # being overlayed, so we should assume its description is OK.
        if "description" in loaded:
            merged._data["description"] = loaded["description"]

        return merged

    @property
    def description(self) -> Optional[str]:
        if "description" in self._data:
            return self._data["description"]

        return None

    @property
    def alignment_holes(self) -> Optional[AlignmentHolesJobSpec]:
        if "alignment_holes" not in self._data:
            return None
        return AlignmentHolesJobSpec(self._data["alignment_holes"])

    @property
    def isolation_routing(self) -> Optional[IsolationRoutingJobSpec]:
        if "isolation_routing" not in self._data:
            return None
        return IsolationRoutingJobSpec(self._data["isolation_routing"])

    @property
    def edge_cuts(self) -> Optional[BoardCutoutJobSpec]:
        if "edge_cuts" not in self._data:
            return None
        return BoardCutoutJobSpec(self._data["edge_cuts"])

    @property
    def drill(self) -> Dict[str, DrillProfileSpec]:
        drill_range_specs = {}

        for name, data in self._data.get("drill", {}).items():
            drill_range_specs[name] = DrillProfileSpec(data, name=name)

        return drill_range_specs

    @property
    def slot(self) -> Dict[str, SlotProfileSpec]:
        drill_range_specs = {}

        for name, data in self._data.get("slot", {}).items():
            drill_range_specs[name] = SlotProfileSpec(data, name=name)

        return drill_range_specs

    def __add__(self, other: Config) -> Config:
        left = copy.deepcopy(self._data)
        right = other._data

        overwrite = ["alignment_holes", "isolation_routing", "edge_cuts"]
        merge = ["drill", "slot"]

        for key in overwrite:
            if key in right:
                left[key] = right[key]

        for key in merge:
            if key in right:
                left.setdefault(key, {}).update(right[key])

        if "description" in left:
            del left["description"]

        return Config(left)


def get_user_config_dir() -> str:
    return os.path.join(appdirs.user_config_dir("barbari", "coddingtonbear"), "configs")


def get_environment_config_file_path() -> str:
    return os.path.join(
        appdirs.user_config_dir("barbari", "coddingtonbear"), "config.yaml"
    )


@dataclass
class EnvironmentConfig:
    flatcam_path: Optional[str] = None
    python_bin: Optional[str] = None


def get_environment_config() -> EnvironmentConfig:
    try:
        with open(get_environment_config_file_path(), "r") as inf:
            loaded = yaml.safe_load(inf)

            return EnvironmentConfig(**loaded)
    except FileNotFoundError:
        return EnvironmentConfig()


def save_environment_config(cfg: EnvironmentConfig) -> None:
    with open(get_environment_config_file_path(), "w") as outf:
        yaml.safe_dump(asdict(cfg), outf)


def get_default_config_dir() -> str:
    return os.path.join(os.path.dirname(__file__), "configs")


def _get_config_path_map() -> Dict[str, str]:
    configs: Dict[str, str] = {}

    directories = [get_default_config_dir(), get_user_config_dir()]

    for directory in directories:
        if not os.path.exists(directory):
            continue

        for filename in os.listdir(directory):
            name, ext = os.path.splitext(filename)

            if ext in (".yaml", ".yml"):
                configs[name] = os.path.join(directory, filename)

    return configs


def get_available_configs() -> List[str]:
    return list(_get_config_path_map().keys())


def get_config_by_name(name: str) -> Config:
    try:
        config_path = _get_config_path_map()[name]
    except KeyError:
        raise exceptions.ConfigNotFound(f"Config '{name}' not found")

    return Config.from_file(config_path)


def get_merged_config(names: List[str]) -> Config:
    configs: List[Config] = []

    for config_name in names:
        configs.append(get_config_by_name(config_name))

    return sum(configs, Config({}))



import logging
import os
import re
from typing import Dict

import gerber

from .constants import LayerType


logger = logging.getLogger(__name__)


class UnknownLayerType(ValueError):
    pass


class GerberProject(object):
    LAYER_NAME_PATTERNS: Dict[LayerType, re.Pattern] = {
        LayerType.B_CU: re.compile(".*\-B(?:[._])Cu\..*"),
        LayerType.F_CU: re.compile(".*\-F(?:[._])Cu\..*"),
        LayerType.EDGE_CUTS: re.compile(".*\-Edge(?:[._])Cuts"),
        LayerType.ALIGNMENT: re.compile(".*-Alignment\..*"),
        LayerType.DRILL: re.compile(".*\.drl$"),
    }

    def __init__(self, path):
        self._path = path
        self._layers = {}

        super().__init__()

    @property
    def path(self) -> str:
        return self._path

    def detect_layer_type(self, filename: str, layer):
        for layer_type, pattern in self.LAYER_NAME_PATTERNS.items():
            if pattern.match(filename):
                return layer_type

        raise UnknownLayerType("Unable to guess layer position for {}".format(filename))

    def get_layers(self):
        if self._layers:
            return self._layers

        for filename in os.listdir(self._path):
            full_path = os.path.join(
                self._path,
                filename,
            )
            if not os.path.isfile(full_path):
                continue

            try:
                layer = gerber.read(full_path)
                layer_type = self.detect_layer_type(filename, layer)
                self._layers[layer_type] = layer
                logger.debug("Loaded %s", full_path)
            except UnknownLayerType:
                logger.error("Could not identify layer type for %s.", full_path)
            except gerber.common.ParseError:
                logger.debug("Unable to parse %s; probably not a gerber.", full_path)

        return self._layers



import argparse

from rich.syntax import Syntax
import yaml

from .. import config
from . import BaseCommand


class Command(BaseCommand):
    FORMAT_HUMAN = "human"
    FORMAT_YAML = "yaml"

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "config",
            nargs="+",
            help="Configuration file to display; later configs override earlier configs -- you can use this to layer your configuration.",
        )
        return super().add_arguments(parser)

    def handle(self) -> None:
        conf = config.get_merged_config(self.options.config)

        self.console.print(Syntax(yaml.safe_dump(conf._data), "yaml"))



import logging
import os
from typing import Callable, Dict, Iterable, List, Mapping, Optional, Union

from gerber import excellon

from .gerbers import GerberProject
from .config import (
    Config,
    DrillHolesJobSpec,
    IsolationRoutingJobSpec,
    JobSpec,
    MillHolesJobSpec,
    MillSlotsJobSpec,
    ToolProfileSpec,
)
from .constants import LayerType, FlatcamLayer


logger = logging.getLogger(__name__)


class FlatcamProcess(object):
    def __init__(self, cmd, *args, **params):
        self._cmd = cmd
        self._args = args
        self._params = params

    def __str__(self):
        cmd_parts = [
            self._cmd,
            *self._args,
            *[
                "-{key} {value}".format(
                    key=key,
                    value=value,
                )
                for key, value in self._params.items()
            ],
        ]

        return " ".join(cmd_parts)

    def get_layer_name(self, layer: Union[FlatcamLayer, str]):
        if isinstance(layer, FlatcamLayer):
            return layer.value

        return layer


class FlatcamCNCJob(FlatcamProcess):
    def __init__(
        self,
        config: JobSpec,
        input_layer: Union[FlatcamLayer, str],
        output_layer: Union[FlatcamLayer, str],
    ):
        extra_kwargs = {}
        if config.multi_depth or config.depth_per_pass:
            extra_kwargs["dpp"] = config.depth_per_pass

        return super().__init__(
            "cncjob",
            self.get_layer_name(input_layer),
            z_cut=config.cut_z,
            z_move=config.travel_z,
            feedrate=config.feed_rate,
            dia=config.tool_size,
            spindlespeed=config.spindle_speed,
            **extra_kwargs,
            outname=self.get_layer_name(output_layer)
        )


class FlatcamDrillCNCJob(FlatcamProcess):
    def __init__(
        self,
        config: DrillHolesJobSpec,
        input_layer: Union[FlatcamLayer, str],
        output_layer: Union[FlatcamLayer, str],
        drilled_dias: List[float],
    ):
        return super().__init__(
            "drillcncjob",
            self.get_layer_name(input_layer),
            drilled_dias=",".join(str(dia) for dia in drilled_dias),
            drillz=config.drill_z,
            travelz=config.travel_z,
            feedrate_z=config.feed_rate,
            spindlespeed=config.spindle_speed,
            outname=self.get_layer_name(output_layer),
        )


class FlatcamMillHoles(FlatcamProcess):
    def __init__(
        self,
        config: MillHolesJobSpec,
        input_layer: Union[FlatcamLayer, str],
        output_layer: Union[FlatcamLayer, str],
        milled_dias: List[float],
    ):
        return super().__init__(
            "milldrills",
            self.get_layer_name(input_layer),
            tooldia=config.tool_size,
            milled_dias=",".join(str(dia) for dia in milled_dias),
            outname=self.get_layer_name(output_layer),
        )


class FlatcamMillSlots(FlatcamProcess):
    def __init__(
        self,
        config: MillSlotsJobSpec,
        input_layer: Union[FlatcamLayer, str],
        output_layer: Union[FlatcamLayer, str],
        milled_dias: List[float],
    ):
        return super().__init__(
            "millslots",
            self.get_layer_name(input_layer),
            tooldia=config.tool_size,
            milled_dias=",".join(str(dia) for dia in milled_dias),
            outname=self.get_layer_name(output_layer),
        )


class FlatcamIsolate(FlatcamProcess):
    def __init__(
        self,
        config: IsolationRoutingJobSpec,
        input_layer: Union[FlatcamLayer, str],
        output_layer: Union[FlatcamLayer, str],
    ):
        return super().__init__(
            "isolate",
            self.get_layer_name(input_layer),
            dia=config.tool_size,
            passes=int(config.passes),
            overlap=config.pass_overlap,
            combine=1,
            outname=self.get_layer_name(output_layer),
        )


class FlatcamWriteGcode(FlatcamProcess):
    def __init__(
        self,
        layer: Union[FlatcamLayer, str],
        path: str,
        counter: int,
        name: str,
        tool_name: str,
        tool_size: float,
    ):
        return super().__init__(
            "write_gcode",
            self.get_layer_name(layer),
            os.path.join(
                path,
                "{counter}.{name}.{tool_size}.{tool_name}.gcode".format(
                    counter=str(counter).zfill(2),
                    name=name,
                    tool_name=tool_name,
                    tool_size=tool_size,
                ),
            ),
        )


class FlatcamProjectGenerator(object):
    def __init__(self, gerbers: GerberProject, config: Config):
        self._gerbers = gerbers
        self._config = config
        self._gcode_counter = 0

        super().__init__()

    @property
    def counter(self) -> int:
        self._gcode_counter += 1
        return self._gcode_counter

    @property
    def counter_str(self) -> str:
        return str(self.counter).zfill(2)

    @property
    def config(self) -> Config:
        return self._config

    @property
    def gerbers(self) -> GerberProject:
        return self._gerbers

    def _load_layers(self) -> Iterable[FlatcamProcess]:
        for layer_type, layer in self.gerbers.get_layers().items():
            if layer_type == LayerType.DRILL:
                yield FlatcamProcess(
                    "open_excellon",
                    layer.filename,
                    outname=layer_type.value,
                )
            else:
                yield FlatcamProcess(
                    "open_gerber",
                    layer.filename,
                    outname=layer_type.value,
                )

    def _alignment_holes(self) -> Iterable[FlatcamProcess]:
        if not self.config.alignment_holes:
            return

        logger.debug("Processing alignment holes...")

        layers = self.gerbers.get_layers()

        edge_cuts = layers[LayerType.EDGE_CUTS]

        min_x = edge_cuts.bounds[0][0]
        max_x = edge_cuts.bounds[0][1]
        min_y = edge_cuts.bounds[1][0]
        max_y = edge_cuts.bounds[1][1]

        hole_offset = (
            self.config.alignment_holes.hole_size / 2
        ) + self.config.alignment_holes.hole_offset
        holes = [
            (
                min_x + hole_offset,
                min_y - hole_offset,
            ),
            (
                max_x - hole_offset,
                min_y - hole_offset,
            ),
        ]

        # This command took some trial and error to figure out --
        # the "holes" parameter is undocumented, and I was only
        # able to figure out how to set the rotation point by
        # reading the source :shrug:
        yield FlatcamProcess(
            "aligndrill",
            FlatcamLayer.EDGE_CUTS.value,
            axis=self.config.alignment_holes.mirror_axis,
            dia=self.config.alignment_holes.hole_size,
            holes='"' + ",".join(str(hole) for hole in holes) + '"',
            dist=max_y / 2,
        )
        yield FlatcamProcess(
            "mirror",
            FlatcamLayer.B_CU.value,
            axis=self.config.alignment_holes.mirror_axis,
            box="edge_cuts",
        )
        yield FlatcamMillHoles(
            self.config.alignment_holes,
            FlatcamLayer.ALIGNMENT,
            FlatcamLayer.ALIGNMENT_PATH,
            milled_dias=[self.config.alignment_holes.hole_size],
        )
        yield FlatcamCNCJob(
            self.config.alignment_holes,
            FlatcamLayer.ALIGNMENT_PATH,
            FlatcamLayer.ALIGNMENT_CNC,
        )
        yield FlatcamWriteGcode(
            FlatcamLayer.ALIGNMENT_CNC,
            self.gerbers.path,
            self.counter,
            "alignment_holes",
            "end_mill",
            self.config.alignment_holes.tool_size,
        )

    def _copper(self) -> Iterable[FlatcamProcess]:
        if not self.config.isolation_routing:
            return

        logger.debug("Processing isolation routing...")

        yield FlatcamIsolate(
            self.config.isolation_routing,
            FlatcamLayer.B_CU,
            FlatcamLayer.B_CU_PATH,
        )
        yield FlatcamCNCJob(
            self.config.isolation_routing,
            FlatcamLayer.B_CU_PATH,
            FlatcamLayer.B_CU_CNC,
        )
        yield FlatcamWriteGcode(
            FlatcamLayer.B_CU_CNC,
            self.gerbers.path,
            self.counter,
            "b_cu",
            "engraving_bit",
            self.config.isolation_routing.tool_size,
        )
        yield FlatcamIsolate(
            self.config.isolation_routing,
            FlatcamLayer.F_CU,
            FlatcamLayer.F_CU_PATH,
        )
        yield FlatcamCNCJob(
            self.config.isolation_routing,
            FlatcamLayer.F_CU_PATH,
            FlatcamLayer.F_CU_CNC,
        )
        yield FlatcamWriteGcode(
            FlatcamLayer.F_CU_CNC,
            self.gerbers.path,
            self.counter,
            "f_cu",
            "engraving_bit",
            self.config.isolation_routing.tool_size,
        )

    def _get_spec_for_tool(
        self, tool, specs: Mapping[str, ToolProfileSpec]
    ) -> Optional[str]:
        selected: Optional[str] = None

        for spec_name, spec in specs.items():
            if spec.allowed_for_tool_size(tool.diameter) and spec.is_better_match_than(
                tool.diameter, specs[selected] if selected else None
            ):
                selected = spec_name

        return selected

    def _get_tool_hit_count(
        self,
        layer: excellon.ExcellonFile,
        tool: excellon.ExcellonTool,
        slot: bool = False,
    ):
        counter = 0

        expected_class = excellon.DrillSlot if slot else excellon.DrillHit

        for hit in layer.hits:
            if isinstance(hit, expected_class) and hit.tool == tool:
                counter += 1

        return counter

    def _drill(self) -> Iterable[FlatcamProcess]:
        if not self.config.drill:
            return

        logger.debug("Processing drills...")

        layer: excellon.ExcellonFile = self.gerbers.get_layers()[LayerType.DRILL]

        process_map: Dict[str, List[int]] = {}
        for tool_number, tool in layer.tools.items():
            if not self._get_tool_hit_count(layer, tool):
                logger.debug(
                    "Tool %s (%s dia) has no drill hits.",
                    tool_number,
                    tool.diameter,
                )
                continue

            selected_spec = self._get_spec_for_tool(tool, self.config.drill)
            if selected_spec:
                logger.debug(
                    "Assigning tool %s (%s dia) to drill process %s.",
                    tool_number,
                    tool.diameter,
                    selected_spec,
                )
                process_map.setdefault(selected_spec, []).append(tool_number)
            else:
                logger.error(
                    "Unable to find compatible drill profile for tool "
                    "#%s having diameter %s; omitting from output.",
                    tool_number,
                    tool.diameter,
                )

        for process_name, tool_numbers in process_map.items():
            specs = self.config.drill[process_name].specs
            for idx, spec in enumerate(specs):
                layer_name = "drill_{name}_{idx}".format(
                    name=process_name,
                    idx=idx,
                )
                if isinstance(spec, DrillHolesJobSpec):
                    yield FlatcamDrillCNCJob(
                        spec,
                        FlatcamLayer.DRILL,
                        layer_name,
                        [layer.tools[n].diameter for n in tool_numbers],
                    )
                    yield FlatcamWriteGcode(
                        layer_name,
                        self.gerbers.path,
                        self.counter,
                        "drill_{name}".format(name=process_name),
                        "drill",
                        spec.tool_size,
                    )
                elif isinstance(spec, MillHolesJobSpec):
                    yield FlatcamMillHoles(
                        spec,
                        FlatcamLayer.DRILL,
                        layer_name,
                        [layer.tools[n].diameter for n in tool_numbers],
                    )
                    yield FlatcamCNCJob(spec, layer_name, layer_name + "_cnc")
                    yield FlatcamWriteGcode(
                        layer_name + "_cnc",
                        self.gerbers.path,
                        self.counter,
                        "drill_{name}".format(name=process_name),
                        "end_mill",
                        spec.tool_size,
                    )
                else:
                    raise ValueError("Unhandled spec!")

    def _slot(self) -> Iterable[FlatcamProcess]:
        if not self.config.slot:
            return

        logger.debug("Processing slots...")

        layer: excellon.ExcellonFile = self.gerbers.get_layers()[LayerType.DRILL]

        process_map: Dict[str, List[int]] = {}
        for tool_number, tool in layer.tools.items():
            if not self._get_tool_hit_count(layer, tool, slot=True):
                logger.debug(
                    "Tool %s (%s dia) has no slot hits.",
                    tool_number,
                    tool.diameter,
                )
                continue

            selected_spec = self._get_spec_for_tool(tool, self.config.slot)
            if selected_spec:
                logger.debug(
                    "Assigning tool %s (%s dia) to slot process %s.",
                    tool_number,
                    tool.diameter,
                    selected_spec,
                )
                process_map.setdefault(selected_spec, []).append(tool_number)
            else:
                logger.error(
                    "Unable to find compatible slot profile for tool "
                    "#%s having diameter %s; omitting from output.",
                    tool_number,
                    tool.diameter,
                )

        for process_name, tool_numbers in process_map.items():
            specs = self.config.slot[process_name].specs
            for idx, spec in enumerate(specs):
                layer_name = "slot_{name}_{idx}".format(
                    name=process_name,
                    idx=idx,
                )
                assert isinstance(spec, MillSlotsJobSpec)

                yield FlatcamMillSlots(
                    spec,
                    FlatcamLayer.DRILL,
                    layer_name,
                    [layer.tools[n].diameter for n in tool_numbers],
                )
                yield FlatcamCNCJob(spec, layer_name, layer_name + "_cnc")
                yield FlatcamWriteGcode(
                    layer_name + "_cnc",
                    self.gerbers.path,
                    self.counter,
                    "slot_{name}".format(name=process_name),
                    "end_mill",
                    spec.tool_size,
                )

    def _edge_cuts(self) -> Iterable[FlatcamProcess]:
        if not self.config.edge_cuts:
            return

        logger.debug("Processing edge cuts...")

        yield FlatcamProcess(
            "cutout",
            FlatcamLayer.EDGE_CUTS.value,
            dia=self.config.edge_cuts.tool_size,
            margin=self.config.edge_cuts.margin,
            gapsize=self.config.edge_cuts.gap_size,
            gaps=self.config.edge_cuts.gaps,
        )
        yield FlatcamCNCJob(
            self.config.edge_cuts,
            FlatcamLayer.EDGE_CUTS_PATH,
            FlatcamLayer.EDGE_CUTS_CNC,
        )
        yield FlatcamWriteGcode(
            FlatcamLayer.EDGE_CUTS_CNC,
            self.gerbers.path,
            self.counter,
            "edge_cuts",
            "end_mill",
            self.config.edge_cuts.tool_size,
        )

    def _quit(self) -> Iterable[FlatcamProcess]:
        yield FlatcamProcess("quit_flatcam")

    def get_cnc_processes(self) -> Iterable[FlatcamProcess]:
        major_step_generators: List[Callable[[], Iterable[FlatcamProcess]]] = [
            self._load_layers,
            self._alignment_holes,
            self._copper,
            self._drill,
            self._slot,
            self._edge_cuts,
            self._quit,
        ]

        for major_step in major_step_generators:
            for step in major_step():
                logger.debug("Step %s generated", step)
                yield step



import argparse
import subprocess

from ..exceptions import BarbariFlatcamError
from .build_script import Command as BuildScriptCommand


class Command(BuildScriptCommand):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--flatcam",
            help="Path to flatcam executable (FlatCAM.py)",
        )
        parser.add_argument(
            "--python-bin",
            help=(
                "Path to the python binary to use when running "
                "FlatCAM.py; set this to the correct python "
                "binary for a virtualenvironment if you are "
                "using one."
            ),
        )
        return super().add_arguments(parser)

    def handle(self) -> None:
        output_file = self.build_script()

        self.console.print(f"Wrote g-code generation script to {output_file}.")

        proc = subprocess.Popen(
            [
                self.options.python_bin or self.config.python_bin or "python",
                self.options.flatcam or self.config.flatcam_path or "./FlatCam.py",
                f"--shellfile={output_file}",
            ],
        )
        result = proc.wait()

        if result != 0:
            raise BarbariFlatcamError(
                "Failed to execute flatcam script; see output above."
            )

        self.console.print("Flatcam executed successfully.")



class BarbariError(Exception):
    pass


class BarbariUserError(BarbariError):
    pass


class ConfigNotFound(BarbariUserError):
    pass


class InvalidConfiguration(BarbariUserError):
    pass


class BarbariFlatcamError(Exception):
    pass



import enum


class LayerType(enum.Enum):
    B_CU = "b_cu"
    F_CU = "f_cu"
    EDGE_CUTS = "edge_cuts"
    DRILL = "drill"
    ALIGNMENT = "alignment"


class FlatcamLayer(enum.Enum):
    B_CU = "b_cu"
    B_CU_PATH = "b_cu_path"
    B_CU_CNC = "b_cu_cnc"
    F_CU = "f_cu"
    F_CU_PATH = "f_cu_path"
    F_CU_CNC = "f_cu_cnc"
    ALIGNMENT = "edge_cuts_aligndrill"
    ALIGNMENT_PATH = "edge_cuts_aligndrill_path"
    ALIGNMENT_CNC = "edge_cuts_aligndrill_cnc"
    EDGE_CUTS = "edge_cuts"
    EDGE_CUTS_PATH = "edge_cuts_cutout"
    EDGE_CUTS_CNC = "edge_cuts_cnc"
    DRILL = "drill"



from __future__ import annotations

from abc import ABCMeta, abstractmethod
import argparse
import logging
import pkg_resources
from typing import Dict, Type

from rich.console import Console

from ..config import EnvironmentConfig, get_environment_config


logger = logging.getLogger(__name__)


def get_installed_commands():
    possible_commands: Dict[str, Type[BaseCommand]] = {}
    for entry_point in pkg_resources.iter_entry_points(group="barbari.commands"):
        try:
            loaded_class = entry_point.load()
        except ImportError:
            logger.warning(
                "Attempted to load entrypoint %s, but " "an ImportError occurred.",
                entry_point,
            )
            continue
        if not issubclass(loaded_class, BaseCommand):
            logger.warning(
                "Loaded entrypoint %s, but loaded class is "
                "not a subclass of `barbari.commands.BaseCommand`.",
                entry_point,
            )
            continue
        possible_commands[entry_point.name] = loaded_class

    return possible_commands


class BaseCommand(metaclass=ABCMeta):
    _options: argparse.Namespace
    _console: Console
    _config: EnvironmentConfig

    def __init__(self, options: argparse.Namespace):
        self._options: argparse.Namespace = options
        self._console = Console()
        self._config = get_environment_config()
        super().__init__()

    @property
    def config(self) -> EnvironmentConfig:
        return self._config

    @property
    def options(self) -> argparse.Namespace:
        return self._options

    @property
    def console(self) -> Console:
        return self._console

    @classmethod
    def get_help(cls) -> str:
        return ""

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        pass

    @classmethod
    def _add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        cls.add_arguments(parser)

    @abstractmethod
    def handle(self) -> None:
        ...



import argparse
import os
import re
from typing import List

from rich.prompt import Confirm

from .. import config, gerbers, flatcam
from . import BaseCommand


class Command(BaseCommand):
    OUTPUT_PATTERN = re.compile("^\d+\..*\.gcode$")

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "directory", help="Path to a directory holding your gerber/drl exports."
        )
        parser.add_argument(
            "config",
            nargs="+",
            help="Configuration file to use; later configs override earlier configs -- you can use this to layer your configuration.",
        )
        return super().add_arguments(parser)

    def get_existing_output(self) -> List[str]:
        existing_files = []

        for filename in os.listdir(
            os.path.abspath(os.path.expanduser(self.options.directory))
        ):
            if self.OUTPUT_PATTERN.match(filename):
                existing_files.append(filename)

        return existing_files

    def build_script(self) -> str:
        project = gerbers.GerberProject(
            os.path.abspath(os.path.expanduser(self.options.directory))
        )
        generator = flatcam.FlatcamProjectGenerator(
            project, config.get_merged_config(self.options.config)
        )

        existing_files = self.get_existing_output()
        if existing_files:
            self.console.print("The following existing flatcam output was found: ")
            for filename in sorted(existing_files):
                self.console.print(f"- {filename}")
            delete_existing = Confirm.ask("Would you like to delete these?")
            if delete_existing:
                for filename in existing_files:
                    os.unlink(
                        os.path.join(
                            os.path.abspath(os.path.expanduser(self.options.directory)),
                            filename,
                        )
                    )

        output_file = os.path.join(
            self.options.directory,
            "generate_gcode.FlatScript",
        )
        processes = generator.get_cnc_processes()
        with open(output_file, "w") as outf:
            for process in processes:
                outf.write(str(process))
                outf.write("\n")

        return output_file

    def handle(self) -> None:
        output_file = self.build_script()

        self.console.print(f"Wrote g-code generation script to {output_file}.")



import argparse
import logging
import sys

from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import install as enable_rich_traceback

from . import exceptions
from .commands import get_installed_commands


logger = logging.getLogger(__name__)


def main(*args):
    enable_rich_traceback()

    commands = get_installed_commands()

    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", default=False, action="store_true")
    parser.add_argument("--verbose", default=False, action="store_true")
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    for cmd_name, cmd_class in commands.items():
        parser_kwargs = {}

        cmd_help = cmd_class.get_help()
        if cmd_help:
            parser_kwargs["help"] = cmd_help

        subparser = subparsers.add_parser(cmd_name, **parser_kwargs)
        cmd_class._add_arguments(subparser)

    args = parser.parse_args()

    logging.basicConfig(
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler()],
        level=logging.DEBUG if args.verbose else logging.INFO,
    )

    if args.debug:
        import debugpy

        debugpy.listen(5678)
        debugpy.wait_for_client()

    console = Console()

    try:
        commands[args.command](args).handle()
    except exceptions.BarbariError as e:
        console.print(f"[red]{e}[/red]")
    except exceptions.BarbariUserError as e:
        console.print(f"[yellow]{e}[/yellow]")
    except Exception:
        console.print_exception()


if __name__ == "__main__":
    main(*sys.argv[1:])
